// Copyright (c) 2010-2011 Zipline Games, Inc. All Rights Reserved.
// http://getmoai.com

#include "pch.h"
#include <Box2D/Box2D.h>
#include <moaicore/MOAIBox2DBody.h>
#include <moaicore/MOAIBox2DFixture.h>
#include <moaicore/MOAILogMessages.h>
#include <moaicore/MOAIBox2DWorld.h>

SUPPRESS_EMPTY_FILE_WARNING
#if USE_BOX2D

//================================================================//
// local
//================================================================//

//----------------------------------------------------------------//
/**	@name	addCircle
	@text	Create and add circle fixture to the body.
	
	@in		MOAIBox2DBody self
	@in		number x
	@in		number y
	@in		number radius
	@out	MOAIBox2DFixture fixture
*/
int MOAIBox2DBody::_addCircle ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "UNNN" )
	float unitsToMeters = self->GetUnitsToMeters ();
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	b2CircleShape circleShape;

	circleShape.m_p.x		= state.GetValue < float >( 2, 0.0f ) * unitsToMeters;
	circleShape.m_p.y		= state.GetValue < float >( 3, 0.0f ) * unitsToMeters;
	circleShape.m_radius	= state.GetValue < float >( 4, 1.0f ) * unitsToMeters;
	
	b2FixtureDef fixtureDef;
	fixtureDef.shape = &circleShape;
	
	MOAIBox2DFixture* fixture = new MOAIBox2DFixture ();
	fixture->SetFixture ( self->mBody->CreateFixture ( &fixtureDef ));
	fixture->SetWorld ( self->mWorld );
	self->mWorld->LuaRetain ( *fixture );

	fixture->PushLuaUserdata ( state );
	return 1;
}

//----------------------------------------------------------------//
/**	@name	addPolygon
	@text	Create and add a polygon fixture to the body.
	
	@in		MOAIBox2DBody self
	@in		table verts					Array containg vertex coordinate components ( t[1] = x0, t[2] = y0, t[3] = x1, t[4] = y1... )
	@out	MOAIBox2DFixture fixture	Returns nil on failure.
*/
int MOAIBox2DBody::_addPolygon ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "U" )

	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}

	float unitsToMeters = self->GetUnitsToMeters ();

	b2Vec2 verts [ MOAIBox2DFixture::MAX_POLY_VERTS ];
	int numVerts = MOAIBox2DFixture::LoadVerts ( state, 2, verts, MOAIBox2DFixture::MAX_POLY_VERTS, unitsToMeters );
			
	if ( numVerts ) {
		
		b2PolygonShape polyShape;
		polyShape.Set ( verts, numVerts );
		
		b2FixtureDef fixtureDef;
		fixtureDef.shape = &polyShape;
		
		MOAIBox2DFixture* fixture = new MOAIBox2DFixture ();
		fixture->SetFixture ( self->mBody->CreateFixture ( &fixtureDef ));
		fixture->SetWorld ( self->mWorld );
		self->mWorld->LuaRetain ( *fixture );

		fixture->PushLuaUserdata ( state );
		return 1;
	}
	return 0;
}

//----------------------------------------------------------------//
/**	@name	addRect
	@text	Create and add a rect fixture to the body.
	
	@in		MOAIBox2DBody self
	@in		number xMin
	@in		number yMin
	@in		number xMax
	@in		number yMax
	@out	MOAIBox2DFixture fixture
*/
int MOAIBox2DBody::_addRect ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "UNNNN" )
	float unitsToMeters = self->GetUnitsToMeters ();
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	USRect rect = state.GetRect < float >( 2 );
	rect.Bless ();
	
	float hx = rect.Width () * 0.5f * unitsToMeters;
	float hy = rect.Height () * 0.5f * unitsToMeters;
	
	b2Vec2 center;
	center.x = ( rect.mXMin * unitsToMeters ) + hx;
	center.y = ( rect.mYMin * unitsToMeters ) + hy;
	
	b2PolygonShape polygonShape;
	polygonShape.SetAsBox ( hx, hy, center, 0.0f );
	
	b2FixtureDef fixtureDef;
	fixtureDef.shape = &polygonShape;
	
	MOAIBox2DFixture* fixture = new MOAIBox2DFixture ();
	fixture->SetFixture ( self->mBody->CreateFixture ( &fixtureDef ));
	fixture->SetWorld ( self->mWorld );
	self->mWorld->LuaRetain ( *fixture );

	fixture->PushLuaUserdata ( state );
	return 1;
}

//----------------------------------------------------------------//
/**	@name	applyAngularImpulse
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@in		number impulse	Angular impulse in degrees.
	@out	nil
*/
int MOAIBox2DBody::_applyAngularImpulse ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "UN" )
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	float impulse = state.GetValue < float >( 2, 0.0f ) * ( float )D2R;
	self->mBody->ApplyAngularImpulse ( impulse );
	
	return 0;
}

//----------------------------------------------------------------//
/**	@name	applyForce
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@in		number forceX
	@in		number forceY
	@opt	number pointX
	@opt	number pointY
	@out	nil
*/
int MOAIBox2DBody::_applyForce ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "UNN" )
	float unitsToMeters = self->GetUnitsToMeters ();
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	b2Vec2 force;
	force.x = state.GetValue < float >( 2, 0.0f ) * unitsToMeters;
	force.y = state.GetValue < float >( 3, 0.0f ) * unitsToMeters;
	
	b2Vec2 point;
	point.x = state.GetValue < float >( 4, 0.0f ) * unitsToMeters;
	point.y = state.GetValue < float >( 5, 0.0f ) * unitsToMeters;
	
	self->mBody->ApplyForce ( force, point );
	
	return 0;
}

//----------------------------------------------------------------//
/**	@name	applyLinearImpulse
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@in		number impulseX
	@in		number impulseY
	@opt	number pointX
	@opt	number pointY
	@out	nil
*/
int MOAIBox2DBody::_applyLinearImpulse ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "UNN" )
	float unitsToMeters = self->GetUnitsToMeters ();
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	b2Vec2 impulse;
	impulse.x = state.GetValue < float >( 2, 0.0f ) * unitsToMeters;
	impulse.y = state.GetValue < float >( 3, 0.0f ) * unitsToMeters;
	
	b2Vec2 point;
	point.x = state.GetValue < float >( 4, 0.0f ) * unitsToMeters;
	point.y = state.GetValue < float >( 5, 0.0f ) * unitsToMeters;
	
	self->mBody->ApplyLinearImpulse ( impulse, point );
	
	return 0;
}

//----------------------------------------------------------------//
/**	@name	applyTorque
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@in		number torque	Torque in degrees.
	@out	nil
*/
int MOAIBox2DBody::_applyTorque ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "U" )
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	float torque = state.GetValue < float >( 2, 0.0f ) * ( float )D2R;
	self->mBody->ApplyTorque ( torque );
	
	return 0;
}

//----------------------------------------------------------------//
/**	@name	destroy
	@text	Schedule body for destruction.
	
	@in		MOAIBox2DBody self
	@out	nil
*/
int MOAIBox2DBody::_destroy ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "U" )
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	if ( self->mWorld ) {
		self->mWorld->ScheduleDestruction ( *self );
	}
	return 0;
}

//----------------------------------------------------------------//
/**	@name	getAngle
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@out	number angle		Angle in degrees.
*/
int MOAIBox2DBody::_getAngle ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "U" )
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	float angle = self->mBody->GetAngle () * ( float )R2D;
	lua_pushnumber ( state, angle );
	return 1;
}

//----------------------------------------------------------------//
/**	@name	getAngularVelocity
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@out	number omega		Anglular velocity in degrees.
*/
int MOAIBox2DBody::_getAngularVelocity ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "U" )
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	float omega = self->mBody->GetAngularVelocity () * ( float )R2D;
	lua_pushnumber ( state, omega );
	return 1;
}

//----------------------------------------------------------------//
/**	@name	getInertia
	@text   See Box2D documentation.

	@in		MOAIBox2DBody self
	@out	number inertia Calculated inertia (based on last call to resetMassData()).
*/
int MOAIBox2DBody::_getInertia ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "U" )
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	lua_pushnumber ( L, self->mBody->GetInertia ());
	return 1;
}

//----------------------------------------------------------------//
/**	@name	getLinearVelocity
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@out	number velocityX
	@out	number velocityY
*/
int MOAIBox2DBody::_getLinearVelocity ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "U" )
	float unitsToMeters = self->GetUnitsToMeters ();
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	b2Vec2 velocity = self->mBody->GetLinearVelocity ();
	lua_pushnumber ( state, velocity.x / unitsToMeters );
	lua_pushnumber ( state, velocity.y / unitsToMeters );
	
	return 2;
}

//----------------------------------------------------------------//
/**	@name	getLocalCenter
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@out	number centerX
	@out	number centerY
*/
int MOAIBox2DBody::_getLocalCenter ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "U" )
	float unitsToMeters = self->GetUnitsToMeters ();
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	b2Vec2 center = self->mBody->GetLocalCenter ();
	lua_pushnumber ( state, center.x / unitsToMeters );
	lua_pushnumber ( state, center.y / unitsToMeters );
	
	return 2;
}

//----------------------------------------------------------------//
/**	@name	getMass
	@text   See Box2D documentation.

	@in		MOAIBox2DBody self
	@out	number Mass Calculated mass (based on last call to resetMassData()).
*/
int MOAIBox2DBody::_getMass ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "U" )
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	lua_pushnumber ( L, self->mBody->GetMass ());
	return 1;
}

//----------------------------------------------------------------//
/**	@name	getPostion
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@out	number positionX
	@out	number positionY
*/
int MOAIBox2DBody::_getPosition ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "U" )
	float unitsToMeters = self->GetUnitsToMeters ();
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	b2Vec2 position = self->mBody->GetPosition ();
	lua_pushnumber ( state, position.x / unitsToMeters );
	lua_pushnumber ( state, position.y / unitsToMeters );
	
	return 2;
}

//----------------------------------------------------------------//
/**	@name	getWorldCenter
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@out	number worldX
	@out	number worldY
*/
int MOAIBox2DBody::_getWorldCenter ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "U" )
	float unitsToMeters = self->GetUnitsToMeters ();
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	b2Vec2 center = self->mBody->GetWorldCenter ();
	lua_pushnumber ( state, center.x / unitsToMeters );
	lua_pushnumber ( state, center.y / unitsToMeters );
	
	return 2;
}

//----------------------------------------------------------------//
/**	@name	isActive
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@out	boolean isActive
*/
int MOAIBox2DBody::_isActive ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "U" )
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	bool isActive = self->mBody->IsActive ();
	lua_pushboolean ( state, isActive );
	
	return 1;
}

//----------------------------------------------------------------//
/**	@name	isAwake
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@out	boolean isAwake
*/
int MOAIBox2DBody::_isAwake ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "U" )
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	bool isAwake = self->mBody->IsAwake ();
	lua_pushboolean ( state, isAwake );
	
	return 1;
}

//----------------------------------------------------------------//
/**	@name	isBullet
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@out	boolean isBullet
*/
int MOAIBox2DBody::_isBullet ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "U" )
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	bool isBullet = self->mBody->IsBullet ();
	lua_pushboolean ( state, isBullet );
	
	return 1;
}

//----------------------------------------------------------------//
/**	@name	isFixedRotation
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@out	boolean isFixedRotation
*/
int MOAIBox2DBody::_isFixedRotation ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "U" )
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	bool isFixedRotation = self->mBody->IsFixedRotation ();
	lua_pushboolean ( state, isFixedRotation );
	
	return 1;
}

//----------------------------------------------------------------//
/**	@name	resetMassData
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@out	nil
*/
int MOAIBox2DBody::_resetMassData ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "U" )
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	self->mBody->ResetMassData ();
	
	return 0;
}

//----------------------------------------------------------------//
/**	@name	setActive
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@opt	boolean active			Default value is false.
	@out	nil
*/
int MOAIBox2DBody::_setActive ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "U" )

	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}

	bool active = state.GetValue < bool >( 2, false );
	self->mBody->SetActive ( active );
	
	return 0;
}

//----------------------------------------------------------------//
/**	@name	setAngularDamping
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@in		number damping
	@out	nil
*/
int MOAIBox2DBody::_setAngularDamping ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "UN" )
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	float damping = state.GetValue < float >( 2, 0.0f );
	self->mBody->SetAngularDamping ( damping );
	
	return 0;
}

//----------------------------------------------------------------//
/**	@name	setAngularVelocity
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@opt	number omega		Angular velocity in degrees. Default value is 0.
	@out	nil
*/
int MOAIBox2DBody::_setAngularVelocity ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "U" )
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	float omega = state.GetValue < float >( 2, 0.0f ) * ( float )D2R;
	self->mBody->SetAngularVelocity ( omega );
	
	return 0;
}

//----------------------------------------------------------------//
/**	@name	setAwake
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@opt	boolean awake		Default value is true.
	@out	nil
*/
int MOAIBox2DBody::_setAwake ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "U" )
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	bool awake = state.GetValue < bool >( 2, true );
	self->mBody->SetAwake ( awake );
	
	return 0;
}

//----------------------------------------------------------------//
/**	@name	setBullet
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@opt	boolean bullet		Default value is true.
	@out	nil
*/
int MOAIBox2DBody::_setBullet ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "U" )
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	bool bullet = state.GetValue < bool >( 2, true );
	self->mBody->SetBullet ( bullet );
	
	return 0;
}

//----------------------------------------------------------------//
/**	@name	setFixedRotation
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@opt	boolean fixedRotation		Default value is true.
	@out	nil
*/
int MOAIBox2DBody::_setFixedRotation ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "U" )
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	bool fixedRotation = state.GetValue < bool >( 2, true );
	self->mBody->SetFixedRotation ( fixedRotation );
	
	return 0;
}

//----------------------------------------------------------------//
/**	@name	setLinearDamping
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@opt	number damping
	@out	nil
*/
int MOAIBox2DBody::_setLinearDamping ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "UN" )
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	float damping = state.GetValue < float >( 2, 0.0f );
	self->mBody->SetLinearDamping ( damping );
	
	return 0;
}

//----------------------------------------------------------------//
/**	@name	setLinearVelocity
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@opt	number velocityX	Default is 0.
	@opt	number velocityY	Default is 0.
	@out	nil
*/
int MOAIBox2DBody::_setLinearVelocity ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "U" )
	float unitsToMeters = self->GetUnitsToMeters ();
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	b2Vec2 v;
	v.x		= state.GetValue < float >( 2, 0.0f ) * unitsToMeters;
	v.y		= state.GetValue < float >( 3, 0.0f ) * unitsToMeters;
	
	self->mBody->SetLinearVelocity ( v );
	
	return 0;
}

//----------------------------------------------------------------//
/**	@name	setMassData
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@in		number mass
	@opt	number I			Default is previous value for I.
	@opt	number centerX		Default is previous value for centerX.
	@opt	number centerY		Default is previous value for centerY.
	@out	nil
*/
int MOAIBox2DBody::_setMassData ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "UN" )
	float unitsToMeters = self->GetUnitsToMeters ();
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	if ( self->mWorld->IsLocked ()) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DWorld_IsLocked );
		return 0;
	}
	
	b2MassData massData;
	self->mBody->GetMassData ( &massData );
	
	massData.mass		= state.GetValue < float >( 2, massData.mass );
	massData.I			= state.GetValue < float >( 3, massData.I );
	massData.center.x	= state.GetValue < float >( 4, massData.center.x ) * unitsToMeters;
	massData.center.y	= state.GetValue < float >( 5, massData.center.y ) * unitsToMeters;
	
	self->mBody->SetMassData ( &massData );
	
	return 0;
}

//----------------------------------------------------------------//
/**	@name	setTransform
	@text	See Box2D documentation.
	
	@in		MOAIBox2DBody self
	@opt	number positionX		Default is 0.
	@opt	number positionY		Default is 0.
	@opt	number angle			In degrees. Default is 0.
	@out	nil
*/
int MOAIBox2DBody::_setTransform ( lua_State* L ) {
	MOAI_LUA_SETUP ( MOAIBox2DBody, "U" )
	float unitsToMeters = self->GetUnitsToMeters ();
	
	if ( !self->mBody ) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DBody_MissingInstance );
		return 0;
	}
	
	if ( self->mWorld->IsLocked ()) {
		MOAILog ( state, MOAILogMessages::MOAIBox2DWorld_IsLocked );
		return 0;
	}
	
	b2Vec2 position;
	position.x		= state.GetValue < float >( 2, 0.0f ) * unitsToMeters;
	position.y		= state.GetValue < float >( 3, 0.0f ) * unitsToMeters;
	float angle		= state.GetValue < float >( 4, 0.0f ) * ( float )D2R;
	
	self->mBody->SetTransform ( position, angle );
	self->ScheduleUpdate ();
	
	return 0;
}

//================================================================//
// MOAIBox2DBody
//================================================================//

//----------------------------------------------------------------//
void MOAIBox2DBody::Destroy () {

	if ( this->mBody ) {
		b2World* world = this->mWorld->mWorld;
		world->DestroyBody ( this->mBody );
		this->mBody = 0;
	}
}

//----------------------------------------------------------------//
MOAIBox2DBody::MOAIBox2DBody () :
	mBody ( 0 ) {
	
	RTTI_BEGIN
		RTTI_EXTEND ( MOAITransformBase )
	RTTI_END
}

//----------------------------------------------------------------//
MOAIBox2DBody::~MOAIBox2DBody () {

	this->Destroy ();
}

//----------------------------------------------------------------//
void MOAIBox2DBody::OnDepNodeUpdate () {

	if ( this->mBody ) {
		
		b2Transform transform = this->mBody->GetTransform ();
		float scale = 1.0f / this->GetUnitsToMeters ();
		
		float* m = this->mLocalToWorldMtx.m;
		
		m [ USAffine2D::C0_R0 ] = ( float )transform.R.col1.x;
		m [ USAffine2D::C0_R1 ] = ( float )transform.R.col1.y;

		m [ USAffine2D::C1_R0 ] = ( float )transform.R.col2.x;
		m [ USAffine2D::C1_R1 ] = ( float )transform.R.col2.y;

		m [ USAffine2D::C2_R0 ] = ( float )transform.position.x * scale;
		m [ USAffine2D::C2_R1 ] = ( float )transform.position.y * scale;
		
		this->mWorldToLocalMtx.Inverse ( this->mLocalToWorldMtx );
	}
}

//----------------------------------------------------------------//
void MOAIBox2DBody::RegisterLuaClass ( USLuaState& state ) {

	MOAITransformBase::RegisterLuaClass ( state );
	
	state.SetField ( -1, "DYNAMIC",		( u32 )b2_dynamicBody );
	state.SetField ( -1, "KINEMATIC",	( u32 )b2_kinematicBody );
	state.SetField ( -1, "STATIC",		( u32 )b2_staticBody );
}

//----------------------------------------------------------------//
void MOAIBox2DBody::RegisterLuaFuncs ( USLuaState& state ) {
	
	MOAITransformBase::RegisterLuaFuncs ( state );
	
	luaL_Reg regTable [] = {
		{ "addCircle",				_addCircle },
		{ "addPolygon",				_addPolygon },
		{ "addRect",				_addRect },
		{ "applyAngularImpulse",	_applyAngularImpulse },
		{ "applyForce",				_applyForce },
		{ "applyLinearImpulse",		_applyLinearImpulse },
		{ "applyTorque",			_applyTorque },
		{ "destroy",				_destroy },
		{ "getAngle",				_getAngle },
		{ "getAngularVelocity",		_getAngularVelocity },
		{ "getInertia",				_getInertia },
		{ "getLinearVelocity",		_getLinearVelocity },
		{ "getLocalCenter",			_getLocalCenter },
		{ "getMass",				_getMass },
		{ "getPosition",			_getPosition },
		{ "getWorldCenter",			_getWorldCenter },
		{ "isActive",				_isActive },
		{ "isAwake",				_isAwake },
		{ "isBullet",				_isBullet },
		{ "isFixedRotation",		_isFixedRotation },
		{ "resetMassData",			_resetMassData },
		{ "setActive",				_setActive },
		{ "setAngularDamping",		_setAngularDamping },
		{ "setAngularVelocity",		_setAngularVelocity },
		{ "setAwake",				_setAwake },
		{ "setBullet",				_setBullet },
		{ "setFixedRotation",		_setFixedRotation },
		{ "setLinearDamping",		_setLinearDamping },
		{ "setLinearVelocity",		_setLinearVelocity },
		{ "setMassData",			_setMassData },
		{ "setTransform",			_setTransform },
		{ NULL, NULL }
	};
	
	luaL_register ( state, 0, regTable );
}

//----------------------------------------------------------------//
void MOAIBox2DBody::SetBody ( b2Body* body ) {

	this->mBody = body;
	body->SetUserData ( this );
}

#endif